@@ -1,40 +1,63 @@
-from typing import List, Any, Dict, Callable, Union
+from __future__ import annotations
 
+from typing import TYPE_CHECKING, Any, Callable, Dict, List, Union
+
+from ..logger import logger
 from .backend import Backend, SparkBackend
-from .common import Column
-from .context import ProcessorContext, VarsContext, TemplatesContext
+from .context import ProcessorContext, TemplatesContext, VarsContext
 from .funcs import FuncRunner
 from .report import SqlProcessorReporter, StepStatus
 from .step import Step, StepFactory
-from ..logger import logger
+
+if TYPE_CHECKING:
+    from pyspark.sql import SparkSession
+
+    from .common import Column
 
 
 def extract_funcs_from_pyfile(funcs_py_file):
+    import importlib
     import os
     import sys
-    import importlib
+
     sys.path.insert(0, os.path.dirname(funcs_py_file))
     func_mod = importlib.import_module(os.path.basename(funcs_py_file)[:-3])
-    funcs = dict([(func, getattr(func_mod, func)) for func in dir(func_mod) if callable(getattr(func_mod, func))])
+    funcs = {func: getattr(func_mod, func) for func in dir(func_mod) if callable(getattr(func_mod, func))}
     return funcs
 
 
 class SqlProcessor:
-
-    def __init__(self, backend: Union['SparkSession', Backend], sql: str, extra_cols: List[Column] = None, variables: dict = None,
-                 report_hdfs_path: str = None, report_task_id: str = None, report_es_url: str = None, report_es_index_prefix: str = None,
-                 scala_udf_initializer: str = None, templates: dict = None, includes: Dict[str, str] = None):
-        backend = backend if isinstance(backend, (Backend, )) else SparkBackend(spark=backend)
+    def __init__(
+        self,
+        backend: Union[SparkSession, Backend],
+        sql: str,
+        extra_cols: List[Column] = None,
+        variables: dict = None,
+        report_hdfs_path: str = None,
+        report_task_id: str = None,
+        report_es_url: str = None,
+        report_es_index_prefix: str = None,
+        scala_udf_initializer: str = None,
+        templates: dict = None,
+        includes: Dict[str, str] = None,
+    ):
+        backend = backend if isinstance(backend, (Backend,)) else SparkBackend(spark=backend)
         self.backend = backend
         self.sql = sql
 
-        self.reporter = SqlProcessorReporter(report_task_id=report_task_id, report_hdfs_path=report_hdfs_path,
-                                             report_es_url=report_es_url, report_es_index_prefix=report_es_index_prefix)
+        self.reporter = SqlProcessorReporter(
+            report_task_id=report_task_id,
+            report_hdfs_path=report_hdfs_path,
+            report_es_url=report_es_url,
+            report_es_index_prefix=report_es_index_prefix,
+        )
         log_var_tmpl_replace = False
         vars_context = VarsContext(debug_log=log_var_tmpl_replace, vars=variables)
         self.func_runner = FuncRunner.create(self.backend)
         vars_context.init(self.func_runner)
-        self.context = ProcessorContext(vars_context, TemplatesContext(debug_log=log_var_tmpl_replace, templates=templates), extra_cols=extra_cols)
+        self.context = ProcessorContext(
+            vars_context, TemplatesContext(debug_log=log_var_tmpl_replace, templates=templates), extra_cols=extra_cols
+        )
         self.step_factory = StepFactory(self.reporter, self.func_runner)
 
         self.step_list = self.step_factory.create_from_sql(self.sql, includes)
@@ -42,7 +65,7 @@ class SqlProcessor:
         self.backend.init_udfs(scala_udf_initializer=scala_udf_initializer)
 
     @property
-    def variables(self) -> [str, Any]:
+    def variables(self) -> Union[str, Any]:
         return self.context.vars_context.vars
 
     @property
@@ -57,7 +80,7 @@ class SqlProcessor:
         if self.backend.is_spark_backend:
             self.backend.set_spark_configs(configs)
         else:
-            logger.warn(f'ignored set spark configs when backend is of type {type(self.backend)}')
+            logger.warn(f"ignored set spark configs when backend is of type {type(self.backend)}")
 
     def register_funcs_from_pyfile(self, funcs_py_file: str):
         funcs = extract_funcs_from_pyfile(funcs_py_file)
@@ -67,7 +90,7 @@ class SqlProcessor:
         self.func_runner.register_funcs(funcs)
 
     def register_udfs_from_pyfile(self, funcs_py_file: str):
-        logger.info(f'resolving udfs from pyfile {funcs_py_file}')
+        logger.info(f"resolving udfs from pyfile {funcs_py_file}")
         funcs = extract_funcs_from_pyfile(funcs_py_file)
         self.register_udfs(funcs)
 
@@ -83,8 +106,9 @@ class SqlProcessor:
     def run_step(self, step: Step, dry_run: bool):
         try:
             # add meta vars to support step information retrieving in functions
-            self.context.add_vars({'__step__': step})
-            self.context.add_vars({'__context__': self.context})
+            self.context.add_vars({"__step__": step})
+            self.context.add_vars({"__context__": self.context})
+            self.context.add_vars({"__backend__": self.backend})
             if not step.should_run(self.context):
                 self.reporter.collect_report(step, status=StepStatus.SKIPPED)
                 return
@@ -94,10 +118,14 @@ class SqlProcessor:
             self.reporter.collect_report(step, status=StepStatus.SUCCEEDED)
         except Exception as e:
             import traceback
+
             self.reporter.collect_report(step, status=StepStatus.FAILED, message=traceback.format_exc())
-            if '__exception_handler__' in self.variables and str(self.variables['__exception_handler__']).upper() != 'NULL':
-                func_name: str = self.variables['__exception_handler__']
-                func_name = func_name.replace('{', '${')
+            if (
+                "__exception_handler__" in self.variables
+                and str(self.variables["__exception_handler__"]).upper() != "NULL"
+            ):
+                func_name: str = self.variables["__exception_handler__"]
+                func_name = func_name.replace("{", "${")
                 self.func_runner.run_func(func_name, self.context.vars_context)(e)
             else:
                 raise e
